@@ -48,29 +48,30 @@ struct exec_options_s
 };
 
 enum
-  {
-    OPTION_CONSOLE_SOCKET = 1000,
-    OPTION_PID_FILE,
-    OPTION_CWD,
-    OPTION_PRESERVE_FDS
-  };
+{
+  OPTION_CONSOLE_SOCKET = 1000,
+  OPTION_PID_FILE,
+  OPTION_CWD,
+  OPTION_PRESERVE_FDS
+};
 
 static struct exec_options_s exec_options;
 
-static struct argp_option options[] =
-  {
-   {"console-socket", OPTION_CONSOLE_SOCKET, "SOCKET", 0, "path to a socket that will receive the master end of the tty", 0},
-   {"tty", 't', "TTY", OPTION_ARG_OPTIONAL, "allocate a pseudo-TTY", 0},
-   {"process", 'p', "FILE", 0, "path to the process.json", 0},
-   {"cwd", OPTION_CWD, "CWD", 0, "current working directory", 0},
-   {"detach", 'd', 0, 0, "detach the command in the background", 0},
-   {"user", 'u', "USERSPEC", 0, "specify the user in the form UID[:GID]", 0},
-   {"env", 'e', "ENV", 0, "add an environment variable", 0},
-   {"cap", 'c', "CAP", 0, "add a capability", 0},
-   {"pid-file", OPTION_PID_FILE, "FILE", 0, "where to write the PID of the container", 0},
-   {"preserve-fds", OPTION_PRESERVE_FDS, 0, 0, "pass additional FDs to the container", 0},
-   {0,}
-  };
+static struct argp_option options[]
+    = { { "console-socket", OPTION_CONSOLE_SOCKET, "SOCKET", 0,
+          "path to a socket that will receive the ptmx end of the tty", 0 },
+        { "tty", 't', "TTY", OPTION_ARG_OPTIONAL, "allocate a pseudo-TTY", 0 },
+        { "process", 'p', "FILE", 0, "path to the process.json", 0 },
+        { "cwd", OPTION_CWD, "CWD", 0, "current working directory", 0 },
+        { "detach", 'd', 0, 0, "detach the command in the background", 0 },
+        { "user", 'u', "USERSPEC", 0, "specify the user in the form UID[:GID]", 0 },
+        { "env", 'e', "ENV", 0, "add an environment variable", 0 },
+        { "cap", 'c', "CAP", 0, "add a capability", 0 },
+        { "pid-file", OPTION_PID_FILE, "FILE", 0, "where to write the PID of the container", 0 },
+        { "preserve-fds", OPTION_PRESERVE_FDS, 0, 0, "pass additional FDs to the container", 0 },
+        {
+            0,
+        } };
 
 static char args_doc[] = "exec CONTAINER cmd";
 
@@ -96,12 +97,15 @@ append_cap (const char *arg)
   exec_options.cap_size++;
 }
 
-static char **dup_array (char **arr, size_t len)
+static char **
+dup_array (char **arr, size_t len)
 {
   size_t i;
   char **ret;
 
-  ret = xmalloc (sizeof (char *) * (len + 1));
+  ret = malloc (sizeof (char *) * (len + 1));
+  if (ret == NULL)
+    error (EXIT_FAILURE, errno, "cannot allocate memory");
   for (i = 0; i < len; i++)
     ret[i] = xstrdup (arr[i]);
 
@@ -131,7 +135,7 @@ parse_opt (int key, char *arg, struct argp_state *state)
       break;
 
     case 'p':
-        exec_options.process = arg;
+      exec_options.process = arg;
       break;
 
     case 't':
@@ -175,9 +179,7 @@ make_oci_process_user (const char *userspec)
   if (userspec == NULL)
     return NULL;
 
-  u = xmalloc (sizeof (runtime_spec_schema_config_schema_process_user));
-  memset (u, 0, sizeof (runtime_spec_schema_config_schema_process_user));
-
+  u = xmalloc0 (sizeof (runtime_spec_schema_config_schema_process_user));
   errno = 0;
   u->uid = strtol (userspec, &endptr, 10);
   if (errno == ERANGE)
@@ -201,7 +203,9 @@ int
 crun_command_exec (struct crun_global_arguments *global_args, int argc, char **argv, libcrun_error_t *err)
 {
   int first_arg, ret = 0;
-  libcrun_context_t crun_context = {0, };
+  libcrun_context_t crun_context = {
+    0,
+  };
 
   crun_context.preserve_fds = 0;
 
@@ -224,13 +228,11 @@ crun_command_exec (struct crun_global_arguments *global_args, int argc, char **a
     return libcrun_container_exec_process_file (&crun_context, argv[first_arg], exec_options.process, err);
   else
     {
-      runtime_spec_schema_config_schema_process *process = xmalloc (sizeof (*process));
+      runtime_spec_schema_config_schema_process *process = xmalloc0 (sizeof (*process));
       int i;
 
-      memset (process, 0, sizeof (*process));
-
       process->args_len = argc;
-      process->args = xmalloc ((argc + 1) * sizeof (*process->args));
+      process->args = xmalloc0 ((argc + 1) * sizeof (*process->args));
       for (i = 0; i < argc - first_arg; i++)
         process->args[i] = xstrdup (argv[first_arg + i + 1]);
       process->args[i] = NULL;
@@ -242,7 +244,8 @@ crun_command_exec (struct crun_global_arguments *global_args, int argc, char **a
       process->user = make_oci_process_user (exec_options.user);
       if (exec_options.cap_size > 0)
         {
-          runtime_spec_schema_config_schema_process_capabilities *capabilities = xmalloc (sizeof (runtime_spec_schema_config_schema_process_capabilities));
+          runtime_spec_schema_config_schema_process_capabilities *capabilities
+              = xmalloc (sizeof (runtime_spec_schema_config_schema_process_capabilities));
 
           capabilities->effective = exec_options.cap;
           capabilities->effective_len = exec_options.cap_size;
