@@ -41,84 +41,30 @@
 #include <linux/magic.h>
 #include <limits.h>
 
-int
-close_and_reset (int *fd)
-{
-  int ret = 0;
-  if (*fd >= 0)
-    {
-      ret = TEMP_FAILURE_RETRY (close (*fd));
-      if (LIKELY (ret == 0))
-        *fd = -1;
-    }
-  return ret;
-}
+#ifndef RESOLVE_IN_ROOT
+#  define RESOLVE_IN_ROOT 0x10
+#endif
+#ifndef __NR_openat2
+#  define __NR_openat2 437
+#endif
 
-void
-cleanup_freep (void *p)
-{
-  void **pp = (void **) p;
-  free (*pp);
-}
+#define MAX_READLINKS 32
 
-void
-cleanup_filep (FILE **f)
+static int
+syscall_openat2 (int dirfd, const char *path, uint64_t flags, uint64_t mode, uint64_t resolve)
 {
-  FILE *file = *f;
-  if (file)
-    (void) fclose (file);
-}
+  struct openat2_open_how
+  {
+    uint64_t flags;
+    uint64_t mode;
+    uint64_t resolve;
+  } how = {
+    .flags = flags,
+    .mode = mode,
+    .resolve = resolve,
+  };
 
-void
-cleanup_closep (void *p)
-{
-  int *pp = p;
-  if (*pp >= 0)
-    TEMP_FAILURE_RETRY (close (*pp));
-}
-
-void
-cleanup_close_vecp (int **p)
-{
-  int *pp = *p;
-  int i;
-
-  for (i = 0; pp[i] >= 0; i++)
-    TEMP_FAILURE_RETRY (close (pp[i]));
-}
-
-void
-cleanup_dirp (DIR **p)
-{
-  DIR *dir = *p;
-  if (dir)
-    closedir (dir);
-}
-
-void *
-xmalloc (size_t size)
-{
-  void *res = malloc (size);
-  if (UNLIKELY (res == NULL))
-    OOM ();
-  return res;
-}
-
-void *
-xrealloc (void *ptr, size_t size)
-{
-  void *res = realloc (ptr, size);
-  if (UNLIKELY (res == NULL))
-    OOM ();
-  return res;
-}
-
-char *
-argp_mandatory_argument (char *arg, struct argp_state *state)
-{
-  if (arg)
-    return arg;
-  return state->argv[state->next++];
+  return (int) syscall (__NR_openat2, dirfd, path, &how, sizeof (how), 0);
 }
 
 int
@@ -146,27 +92,13 @@ xasprintf (char **str, const char *fmt, ...)
   return ret;
 }
 
-char *
-xstrdup (const char *str)
-{
-  char *ret;
-  if (str == NULL)
-    return NULL;
-
-  ret = strdup (str);
-  if (ret == NULL)
-    OOM ();
-
-  return ret;
-}
-
 int
 write_file_at (int dirfd, const char *name, const void *data, size_t len, libcrun_error_t *err)
 {
   cleanup_close int fd = openat (dirfd, name, O_WRONLY | O_CREAT, 0700);
   int ret = 0;
   if (UNLIKELY (fd < 0))
-    return crun_make_error (err, errno, "writing file `%s`", name);
+    return crun_make_error (err, errno, "opening file `%s` for writing", name);
 
   if (len)
     {
@@ -184,7 +116,7 @@ write_file (const char *name, const void *data, size_t len, libcrun_error_t *err
   cleanup_close int fd = open (name, O_WRONLY | O_CREAT, 0700);
   int ret;
   if (UNLIKELY (fd < 0))
-    return crun_make_error (err, errno, "writing file `%s`", name);
+    return crun_make_error (err, errno, "opening file `%s` for writing", name);
 
   ret = TEMP_FAILURE_RETRY (write (fd, data, len));
   if (UNLIKELY (ret < 0))
@@ -227,7 +159,7 @@ get_file_type_fd (int fd, mode_t *mode)
   *mode = stx.stx_mode;
   return ret;
 
- fallback:
+fallback:
 #endif
   ret = fstat (fd, &st);
   *mode = st.st_mode;
@@ -254,7 +186,7 @@ get_file_type_at (int dirfd, mode_t *mode, bool nofollow, const char *path)
   *mode = stx.stx_mode;
   return ret;
 
- fallback:
+fallback:
 #endif
   ret = fstatat (dirfd, path, &st, nofollow ? AT_SYMLINK_NOFOLLOW : 0);
   *mode = st.st_mode;
@@ -270,7 +202,7 @@ get_file_type (mode_t *mode, bool nofollow, const char *path)
 int
 create_file_if_missing_at (int dirfd, const char *file, libcrun_error_t *err)
 {
-  cleanup_close int fd_write = openat (dirfd, file, O_CREAT | O_WRONLY, 0700);
+  cleanup_close int fd_write = openat (dirfd, file, O_CLOEXEC | O_CREAT | O_WRONLY, 0700);
   if (fd_write < 0)
     {
       mode_t mode;
@@ -286,25 +218,6 @@ create_file_if_missing_at (int dirfd, const char *file, libcrun_error_t *err)
   return 0;
 }
 
-int
-create_file_if_missing (const char *file, libcrun_error_t *err)
-{
-  cleanup_close int fd_write = open (file, O_CLOEXEC | O_CREAT | O_WRONLY, 0700);
-  if (fd_write < 0)
-    {
-      mode_t mode;
-      int ret;
-
-      /* On errors, check if the file already exists.  */
-      ret = get_file_type (&mode, false, file);
-      if (ret == 0 && S_ISREG (mode))
-        return 0;
-
-      return crun_make_error (err, errno, "creating file `%s`", file);
-    }
-  return 0;
-}
-
 static int
 ensure_directory_internal_at (int dirfd, char *path, size_t len, int mode, libcrun_error_t *err)
 {
@@ -375,7 +288,7 @@ crun_ensure_directory_at (int dirfd, const char *path, int mode, bool nofollow,
   return 0;
 }
 
-int
+static int
 check_fd_under_path (const char *rootfs, size_t rootfslen, int fd, const char *fdname, libcrun_error_t *err)
 {
   int ret;
@@ -403,19 +316,78 @@ close_and_replace (int *oldfd, int newfd)
   *oldfd = newfd;
 }
 
+/* Defined in chroot_realpath.c  */
+char *chroot_realpath (const char *chroot, const char *path, char resolved_path[]);
+
+int
+safe_openat (int dirfd, const char *rootfs, size_t rootfs_len, const char *path, int flags, int mode,
+             libcrun_error_t *err)
+{
+  int ret;
+  cleanup_close int fd = -1;
+  static bool openat2_supported = true;
+  char buffer[PATH_MAX], *path_in_chroot;
+
+  if (openat2_supported)
+    {
+      ret = syscall_openat2 (dirfd, path, flags, mode, RESOLVE_IN_ROOT);
+      if (ret < 0)
+        {
+          if (errno == ENOSYS)
+            openat2_supported = false;
+          if (errno == ENOSYS || errno == EINVAL)
+            goto fallback;
+          return crun_make_error (err, errno, "openat2 `%s`", path);
+        }
+
+      return ret;
+    }
+
+fallback:
+  path_in_chroot = chroot_realpath (rootfs, path, buffer);
+  if (path_in_chroot == NULL)
+    return crun_make_error (err, errno, "cannot resolve `%s` under rootfs", path);
+
+  path_in_chroot += rootfs_len;
+  while (*path_in_chroot == '/')
+    path_in_chroot++;
+
+  ret = openat (dirfd, path_in_chroot, flags, mode);
+  if (UNLIKELY (ret < 0))
+    return crun_make_error (err, errno, "open `%s`", path);
+
+  fd = ret;
+
+  ret = check_fd_under_path (rootfs, rootfs_len, fd, path, err);
+  if (UNLIKELY (ret < 0))
+    return ret;
+
+  ret = fd;
+  fd = -1;
+  return ret;
+}
+
 static int
-crun_safe_ensure_at (bool dir, int dirfd, const char *dirpath, size_t dirpath_len, const char *path,
-                     int mode, libcrun_error_t *err)
+crun_safe_ensure_at (bool dir, int dirfd, const char *dirpath, size_t dirpath_len, const char *path, int mode,
+                     int max_readlinks, libcrun_error_t *err)
 {
-  cleanup_free char *npath = xstrdup (path);
-  cleanup_close int wd_file_cleanup = -1;
   cleanup_close int wd_cleanup = -1;
+  cleanup_free char *npath = NULL;
   bool last_component = false;
+  size_t depth = 0;
   const char *cur;
   char *it;
   int cwd;
   int ret;
 
+  if (max_readlinks <= 0)
+    return crun_make_error (err, ELOOP, "resolve path `%s`", path);
+
+  while (*path == '/')
+    path++;
+
+  npath = xstrdup (path);
+
   it = npath + strlen (npath) - 1;
   while (*it == '/' && it > npath && ((size_t) (it - path)) > dirpath_len)
     *it-- = '\0';
@@ -435,24 +407,60 @@ crun_safe_ensure_at (bool dir, int dirfd, const char *dirpath, size_t dirpath_le
       if (cur[0] == '\0')
         break;
 
-      if (!last_component || dir)
+      if (strcmp (cur, ".") == 0)
+        goto next;
+      else if (strcmp (cur, ".."))
+        depth++;
+      else
+        {
+          if (depth)
+            depth--;
+          else
+            {
+              close_and_reset (&wd_cleanup);
+              cwd = dirfd;
+            }
+        }
+
+      if (! last_component || dir)
         ret = mkdirat (cwd, cur, mode);
       else
         {
-          ret = openat (cwd, cur, O_CLOEXEC | O_CREAT | O_WRONLY, 0700);
+          ret = openat (cwd, cur, O_CLOEXEC | O_CREAT | O_WRONLY | O_NOFOLLOW, 0700);
           if (UNLIKELY (ret < 0))
             {
-              int saved_errno = errno;
+              /* If the last component is a symlink, repeat the lookup with the resolved path.  */
+              if (errno == ELOOP)
+                {
+                  size_t s, resolved_size = 0;
+                  cleanup_free char *resolved_path = NULL;
 
+                  do
+                    {
+                      resolved_size += 512;
+                      resolved_path = xrealloc (resolved_path, resolved_size);
+
+                      s = readlinkat (dirfd, npath, resolved_path, resolved_size);
+                    }
+                  while (s == resolved_size);
+
+                  if (s > 0)
+                    {
+                      resolved_path[s] = '\0';
+                      crun_error_release (err);
+                      return crun_safe_ensure_at (dir, dirfd, dirpath, dirpath_len, resolved_path, mode,
+                                                  max_readlinks - 1, err);
+                    }
+                }
               /* If the previous openat fails, attempt to open the file in O_PATH mode.  */
-              ret = openat (cwd, cur, O_CLOEXEC | O_PATH);
+              ret = openat (cwd, cur, O_CLOEXEC | O_PATH, 0);
               if (ret < 0)
-                return crun_make_error (err, saved_errno, "create `%s`", path);
+                return crun_make_error (err, errno, "open `%s/%s`", dirpath, cur);
             }
 
           close_and_replace (&wd_cleanup, ret);
 
-          return check_fd_under_path (dirpath, dirpath_len, ret, path, err);
+          return 0;
         }
 
       if (ret < 0)
@@ -461,20 +469,18 @@ crun_safe_ensure_at (bool dir, int dirfd, const char *dirpath, size_t dirpath_le
             return crun_make_error (err, errno, "mkdir `%s`", cur);
         }
 
-      cwd = openat (cwd, cur, O_CLOEXEC | O_PATH);
+      cwd = safe_openat (dirfd, dirpath, dirpath_len, npath, O_CLOEXEC | O_PATH, 0, err);
       if (UNLIKELY (cwd < 0))
-        return crun_make_error (err, errno, "open `%s`", cur);
+        return cwd;
 
       close_and_replace (&wd_cleanup, cwd);
 
-      ret = check_fd_under_path (dirpath, dirpath_len, cwd, cur, err);
-      if (UNLIKELY (ret < 0))
-        return ret;
-
+    next:
       if (it == NULL)
         break;
 
       cur = it + 1;
+      *it = '/';
       it = strchr (cur, '/');
     }
 
@@ -485,14 +491,14 @@ int
 crun_safe_ensure_directory_at (int dirfd, const char *dirpath, size_t dirpath_len, const char *path, int mode,
                                libcrun_error_t *err)
 {
-  return crun_safe_ensure_at (true, dirfd, dirpath, dirpath_len, path, mode, err);
+  return crun_safe_ensure_at (true, dirfd, dirpath, dirpath_len, path, mode, MAX_READLINKS, err);
 }
 
 int
 crun_safe_ensure_file_at (int dirfd, const char *dirpath, size_t dirpath_len, const char *path, int mode,
                           libcrun_error_t *err)
 {
-  return crun_safe_ensure_at (false, dirfd, dirpath, dirpath_len, path, mode, err);
+  return crun_safe_ensure_at (false, dirfd, dirpath, dirpath_len, path, mode, MAX_READLINKS, err);
 }
 
 int
@@ -549,7 +555,7 @@ get_file_size (int fd, off_t *size)
 
   return ret;
 
- fallback:
+fallback:
 #endif
   ret = fstat (fd, &st);
   *size = st.st_size;
@@ -615,7 +621,7 @@ libcrun_initialize_selinux (libcrun_error_t *err)
 }
 
 int
-libcrun_initialize_apparmor (libcrun_error_t *err arg_unused)
+libcrun_initialize_apparmor (libcrun_error_t *err)
 {
   cleanup_close int fd = -1;
   int size;
@@ -624,13 +630,16 @@ libcrun_initialize_apparmor (libcrun_error_t *err arg_unused)
   if (apparmor_enabled >= 0)
     return apparmor_enabled;
 
-  fd = open ("/sys/module/apparmor/parameters/enabled", O_RDONLY | O_CLOEXEC);
-  if (fd == -1)
-    return 0;
+  if (crun_dir_p_at (AT_FDCWD, "/sys/kernel/security/apparmor", true, err))
+    {
+      fd = open ("/sys/module/apparmor/parameters/enabled", O_RDONLY | O_CLOEXEC);
+      if (fd == -1)
+        return 0;
 
-  size = TEMP_FAILURE_RETRY (read (fd, &buf, 2));
+      size = TEMP_FAILURE_RETRY (read (fd, &buf, 2));
 
-  apparmor_enabled = size > 0 && buf[0] == 'Y' ? 1 : 0;
+      apparmor_enabled = size > 0 && buf[0] == 'Y' ? 1 : 0;
+    }
 
   return apparmor_enabled;
 }
@@ -662,11 +671,11 @@ add_selinux_mount_label (char **retlabel, const char *data, const char *label, l
     }
   *retlabel = xstrdup (data);
   return 0;
-
 }
 
 static int
-write_file_and_check_fs_type (const char *file, const char *data, size_t len, unsigned int type, const char *type_name, libcrun_error_t *err)
+write_file_and_check_fs_type (const char *file, const char *data, size_t len, unsigned int type, const char *type_name,
+                              libcrun_error_t *err)
 {
   int ret;
   struct statfs sfs;
@@ -701,7 +710,8 @@ set_selinux_exec_label (const char *label, libcrun_error_t *err)
 
   if (ret)
     {
-      ret = write_file_and_check_fs_type ("/proc/thread-self/attr/exec", label, strlen (label), PROC_SUPER_MAGIC, "procfs", err);
+      ret = write_file_and_check_fs_type ("/proc/thread-self/attr/exec", label, strlen (label), PROC_SUPER_MAGIC,
+                                          "procfs", err);
       if (UNLIKELY (ret < 0))
         return ret;
     }
@@ -730,7 +740,8 @@ set_apparmor_profile (const char *profile, libcrun_error_t *err)
 
       xasprintf (&buf, "exec %s", profile);
 
-      ret = write_file_and_check_fs_type ("/proc/thread-self/attr/exec", buf, strlen (buf), PROC_SUPER_MAGIC, "procfs", err);
+      ret = write_file_and_check_fs_type ("/proc/thread-self/attr/exec", buf, strlen (buf), PROC_SUPER_MAGIC, "procfs",
+                                          err);
       if (UNLIKELY (ret < 0))
         return ret;
     }
@@ -752,7 +763,7 @@ read_all_fd (int fd, const char *description, char **out, size_t *len, libcrun_e
   /* NUL terminate the buffer.  */
   allocated = size;
   if (size == 0)
-    allocated = 256;
+    allocated = 4096;
   buf = xmalloc (allocated + 1);
   nread = 0;
   while ((size && nread < (size_t) size) || size == 0)
@@ -766,7 +777,7 @@ read_all_fd (int fd, const char *description, char **out, size_t *len, libcrun_e
 
       nread += ret;
 
-      allocated += 256;
+      allocated += 4096;
       buf = xrealloc (buf, allocated + 1);
     }
   buf[nread] = '\0';
@@ -778,30 +789,35 @@ read_all_fd (int fd, const char *description, char **out, size_t *len, libcrun_e
 }
 
 int
-read_all_file (const char *path, char **out, size_t *len, libcrun_error_t *err)
+read_all_file_at (int dirfd, const char *path, char **out, size_t *len, libcrun_error_t *err)
 {
   cleanup_close int fd;
 
-  if (strcmp (path, "-") == 0)
-    path = "/dev/stdin";
-
-  fd = TEMP_FAILURE_RETRY (open (path, O_RDONLY));
+  fd = TEMP_FAILURE_RETRY (openat (dirfd, path, O_RDONLY));
   if (UNLIKELY (fd < 0))
     return crun_make_error (err, errno, "error opening file `%s`", path);
 
   return read_all_fd (fd, path, out, len, err);
 }
 
+int
+read_all_file (const char *path, char **out, size_t *len, libcrun_error_t *err)
+{
+  if (strcmp (path, "-") == 0)
+    path = "/dev/stdin";
+
+  return read_all_file_at (AT_FDCWD, path, out, len, err);
+}
+
 int
 open_unix_domain_client_socket (const char *path, int dgram, libcrun_error_t *err)
 {
-  struct sockaddr_un addr;
+  struct sockaddr_un addr = {};
   int ret;
   cleanup_close int fd = socket (AF_UNIX, dgram ? SOCK_DGRAM : SOCK_STREAM, 0);
   if (UNLIKELY (fd < 0))
     return crun_make_error (err, errno, "error creating UNIX socket");
 
-  memset (&addr, 0, sizeof (addr));
   if (strlen (path) >= sizeof (addr.sun_path))
     return crun_make_error (err, 0, "invalid path %s specified", path);
   strcpy (addr.sun_path, path);
@@ -819,13 +835,12 @@ open_unix_domain_client_socket (const char *path, int dgram, libcrun_error_t *er
 int
 open_unix_domain_socket (const char *path, int dgram, libcrun_error_t *err)
 {
-  struct sockaddr_un addr;
+  struct sockaddr_un addr = {};
   int ret;
   cleanup_close int fd = socket (AF_UNIX, dgram ? SOCK_DGRAM : SOCK_STREAM, 0);
   if (UNLIKELY (fd < 0))
     return crun_make_error (err, errno, "error creating UNIX socket");
 
-  memset (&addr, 0, sizeof (addr));
   if (strlen (path) >= sizeof (addr.sun_path))
     return crun_make_error (err, 0, "invalid path %s specified", path);
   strcpy (addr.sun_path, path);
@@ -834,7 +849,7 @@ open_unix_domain_socket (const char *path, int dgram, libcrun_error_t *err)
   if (UNLIKELY (ret < 0))
     return crun_make_error (err, errno, "bind socket to `%s`", path);
 
-  if (!dgram)
+  if (! dgram)
     {
       ret = listen (fd, 1);
       if (UNLIKELY (ret < 0))
@@ -853,13 +868,10 @@ send_fd_to_socket (int server, int fd, libcrun_error_t *err)
   int ret;
   struct cmsghdr *cmsg = NULL;
   struct iovec iov[1];
-  struct msghdr msg;
-  char ctrl_buf[CMSG_SPACE (sizeof (int))];
+  struct msghdr msg = {};
+  char ctrl_buf[CMSG_SPACE (sizeof (int))] = {};
   char data[1];
 
-  memset (&msg, 0, sizeof (struct msghdr));
-  memset (ctrl_buf, 0, CMSG_SPACE (sizeof (int)));
-
   data[0] = ' ';
   iov[0].iov_base = data;
   iov[0].iov_len = sizeof (data);
@@ -890,14 +902,11 @@ receive_fd_from_socket (int from, libcrun_error_t *err)
   cleanup_close int fd = -1;
   int ret;
   struct iovec iov[1];
-  struct msghdr msg;
-  char ctrl_buf[CMSG_SPACE (sizeof (int))];
+  struct msghdr msg = {};
+  char ctrl_buf[CMSG_SPACE (sizeof (int))] = {};
   char data[1];
   struct cmsghdr *cmsg;
 
-  memset (&msg, 0, sizeof (struct msghdr));
-  memset (ctrl_buf, 0, CMSG_SPACE (sizeof (int)));
-
   data[0] = ' ';
   iov[0].iov_base = data;
   iov[0].iov_len = sizeof (data);
@@ -998,7 +1007,7 @@ copy_from_fd_to_fd (int src, int dst, int consume, libcrun_error_t *err)
 
     fallback:
 #endif
-# define BUFFER_SIZE 4096
+#define BUFFER_SIZE 4096
 
       buffer = xmalloc (BUFFER_SIZE);
       nread = TEMP_FAILURE_RETRY (read (src, buffer, BUFFER_SIZE));
@@ -1021,7 +1030,6 @@ copy_from_fd_to_fd (int src, int dst, int consume, libcrun_error_t *err)
   while (consume && nread);
 
   return 0;
-
 }
 
 int
@@ -1044,10 +1052,77 @@ run_process (char **args, libcrun_error_t *err)
   _exit (EXIT_FAILURE);
 }
 
+#ifndef HAVE_FGETPWENT_R
+static unsigned
+atou (char **s)
+{
+  unsigned x;
+  for (x = 0; **s - '0' < 10; ++*s)
+    x = 10 * x + (**s - '0');
+  return x;
+}
+
+int
+fgetpwent_r (FILE *f, struct passwd *pw, char *line, size_t size, struct passwd **res)
+{
+  char *s;
+  int rv = 0;
+  for (;;)
+    {
+      line[size - 1] = '\xff';
+      if ((fgets (line, size, f) == NULL) || ferror (f) || line[size - 1] != '\xff')
+        {
+          rv = (line[size - 1] != '\xff') ? ERANGE : ENOENT;
+          line = 0;
+          pw = 0;
+          break;
+        }
+      line[strcspn (line, "\n")] = 0;
+
+      s = line;
+      pw->pw_name = s++;
+      if (! (s = strchr (s, ':')))
+        continue;
+
+      *s++ = 0;
+      pw->pw_passwd = s;
+      if (! (s = strchr (s, ':')))
+        continue;
+
+      *s++ = 0;
+      pw->pw_uid = atou (&s);
+      if (*s != ':')
+        continue;
+
+      *s++ = 0;
+      pw->pw_gid = atou (&s);
+      if (*s != ':')
+        continue;
+
+      *s++ = 0;
+      pw->pw_gecos = s;
+      if (! (s = strchr (s, ':')))
+        continue;
+
+      *s++ = 0;
+      pw->pw_dir = s;
+      if (! (s = strchr (s, ':')))
+        continue;
+
+      *s++ = 0;
+      pw->pw_shell = s;
+      break;
+    }
+  *res = pw;
+  if (rv)
+    errno = rv;
+  return rv;
+}
+#endif
+
 int
 set_home_env (uid_t id)
 {
-#ifdef HAVE_FGETPWENT_R
   struct passwd pwd;
   cleanup_free char *buf = NULL;
   long buf_size;
@@ -1093,11 +1168,6 @@ set_home_env (uid_t id)
         }
     }
   return 0;
-#else
-  (void) id;
-  errno = ENOTSUP;
-  return -1;
-#endif
 }
 
 /*if subuid or subgid exist, take the first range for the user */
@@ -1170,7 +1240,7 @@ getsubidrange (uid_t id, int is_uid, uint32_t *from, uint32_t *len)
     }
 }
 
-#define MIN(x,y) ((x)<(y)?(x):(y))
+#define MIN(x, y) ((x) < (y) ? (x) : (y))
 
 size_t
 format_default_id_mapping (char **ret, uid_t container_id, uid_t host_id, int is_uid)
@@ -1209,14 +1279,8 @@ format_default_id_mapping (char **ret, uid_t container_id, uid_t host_id, int is
 
 /* will leave SIGCHLD blocked if TIMEOUT is used.  */
 int
-run_process_with_stdin_timeout_envp (char *path,
-                                     char **args,
-                                     const char *cwd,
-                                     int timeout,
-                                     char **envp,
-                                     char *stdin,
-                                     size_t stdin_len,
-                                     libcrun_error_t *err)
+run_process_with_stdin_timeout_envp (char *path, char **args, const char *cwd, int timeout, char **envp, char *stdin,
+                                     size_t stdin_len, int out_fd, int err_fd, libcrun_error_t *err)
 {
   int stdin_pipe[2];
   pid_t pid;
@@ -1275,12 +1339,12 @@ run_process_with_stdin_timeout_envp (char *path,
               if (ret < 0 && errno == EAGAIN)
                 goto timeout;
             }
- timeout:
+        timeout:
           kill (pid, SIGKILL);
           return crun_make_error (err, 0, "timeout expired for `%s`", path);
         }
 
- read_waitpid:
+    read_waitpid:
       r = TEMP_FAILURE_RETRY (waitpid (pid, &status, 0));
       if (r < 0)
         return crun_make_error (err, errno, "waitpid");
@@ -1291,19 +1355,29 @@ run_process_with_stdin_timeout_envp (char *path,
     }
   else
     {
-      char *tmp_args[] = {path, NULL};
-      int out = open ("/dev/null", O_WRONLY);
+      char *tmp_args[] = { path, NULL };
+      int dev_null_fd = -1;
 
-      if (UNLIKELY (out < 0))
-        _exit (EXIT_FAILURE);
+      if (out_fd < 0 || err_fd < 0)
+        {
+          dev_null_fd = open ("/dev/null", O_WRONLY);
+          if (UNLIKELY (dev_null_fd < 0))
+            _exit (EXIT_FAILURE);
+        }
 
       TEMP_FAILURE_RETRY (close (pipe_w));
       dup2 (pipe_r, 0);
       TEMP_FAILURE_RETRY (close (pipe_r));
 
-      dup2 (out, 1);
-      dup2 (out, 2);
-      TEMP_FAILURE_RETRY (close (out));
+      dup2 (out_fd >= 0 ? out_fd : dev_null_fd, 1);
+      dup2 (err_fd >= 0 ? err_fd : dev_null_fd, 2);
+
+      if (dev_null_fd >= 0)
+        TEMP_FAILURE_RETRY (close (dev_null_fd));
+      if (out_fd >= 0)
+        TEMP_FAILURE_RETRY (close (out_fd));
+      if (err_fd >= 0)
+        TEMP_FAILURE_RETRY (close (err_fd));
 
       if (args == NULL)
         args = tmp_args;
@@ -1394,15 +1468,15 @@ set_blocking_fd (int fd, int blocking, libcrun_error_t *err)
 int
 parse_json_file (yajl_val *out, const char *jsondata, struct parser_context *ctx arg_unused, libcrun_error_t *err)
 {
-    char errbuf[1024];
+  char errbuf[1024];
 
-    *err = NULL;
+  *err = NULL;
 
-    *out = yajl_tree_parse (jsondata, errbuf, sizeof (errbuf));
-    if (*out == NULL)
-      return crun_make_error (err, 0, "cannot parse the data: `%s`", errbuf);
+  *out = yajl_tree_parse (jsondata, errbuf, sizeof (errbuf));
+  if (*out == NULL)
+    return crun_make_error (err, 0, "cannot parse the data: `%s`", errbuf);
 
-    return 0;
+  return 0;
 }
 
 int
@@ -1451,7 +1525,7 @@ find_executable (const char *executable_path, const char *cwd)
       return NULL;
     }
 
-  if (executable_path[0] == '.')
+  if (executable_path[0] == '.' || (executable_path[0] != '/' && strchr (executable_path, '/')))
     {
       cleanup_free char *cwd_allocated = NULL;
 
@@ -1559,7 +1633,6 @@ copy_xattr (int sfd, int dfd, const char *srcname, const char *destname, libcrun
       return crun_make_error (err, errno, "get xattr list for `%s`", srcname);
     }
 
-
   if (xattr_len == 0)
     return 0;
 
@@ -1593,8 +1666,8 @@ copy_xattr (int sfd, int dfd, const char *srcname, const char *destname, libcrun
 
 #endif
 
-static
-int copy_rec_stat_file_at (int dfd, const char *path, mode_t *mode, off_t *size, dev_t *rdev, uid_t *uid, gid_t *gid)
+static int
+copy_rec_stat_file_at (int dfd, const char *path, mode_t *mode, off_t *size, dev_t *rdev, uid_t *uid, gid_t *gid)
 {
   struct stat st;
   int ret;
@@ -1602,7 +1675,8 @@ int copy_rec_stat_file_at (int dfd, const char *path, mode_t *mode, off_t *size,
 #ifdef HAVE_STATX
   struct statx stx;
 
-  ret = statx (dfd, path, AT_SYMLINK_NOFOLLOW | AT_STATX_DONT_SYNC, STATX_TYPE | STATX_MODE | STATX_SIZE | STATX_UID | STATX_GID, &stx);
+  ret = statx (dfd, path, AT_SYMLINK_NOFOLLOW | AT_STATX_DONT_SYNC,
+               STATX_TYPE | STATX_MODE | STATX_SIZE | STATX_UID | STATX_GID, &stx);
   if (UNLIKELY (ret < 0))
     {
       if (errno == ENOSYS || errno == EINVAL)
@@ -1619,7 +1693,7 @@ int copy_rec_stat_file_at (int dfd, const char *path, mode_t *mode, off_t *size,
 
   return ret;
 
- fallback:
+fallback:
 #endif
   ret = fstatat (dfd, path, &st, AT_SYMLINK_NOFOLLOW);
 
@@ -1633,9 +1707,9 @@ int copy_rec_stat_file_at (int dfd, const char *path, mode_t *mode, off_t *size,
 }
 
 int
-copy_recursive_fd_to_fd (int srcdirfd, int destdirfd, const char *srcname, const char *destname, libcrun_error_t *err)
+copy_recursive_fd_to_fd (int srcdirfd, int dfd, const char *srcname, const char *destname, libcrun_error_t *err)
 {
-  cleanup_close int dfd = destdirfd;
+  cleanup_close int destdirfd = dfd;
   cleanup_dir DIR *dsrcfd = NULL;
   struct dirent *de;
 
@@ -1660,8 +1734,7 @@ copy_recursive_fd_to_fd (int srcdirfd, int destdirfd, const char *srcname, const
       uid_t uid;
       gid_t gid;
 
-      if (strcmp (de->d_name, ".") == 0 ||
-          strcmp (de->d_name, "..") == 0)
+      if (strcmp (de->d_name, ".") == 0 || strcmp (de->d_name, "..") == 0)
         continue;
 
       ret = copy_rec_stat_file_at (dirfd (dsrcfd), de->d_name, &mode, &st_size, &rdev, &uid, &gid);
@@ -1753,11 +1826,11 @@ copy_recursive_fd_to_fd (int srcdirfd, int destdirfd, const char *srcname, const
       if (UNLIKELY (ret < 0))
         return crun_make_error (err, errno, "chown `%s/%s`", destname, de->d_name);
 
-      /*
-       * ALLPERMS is not defined by POSIX
-       */
+        /*
+         * ALLPERMS is not defined by POSIX
+         */
 #ifndef ALLPERMS
-# define ALLPERMS (S_ISUID|S_ISGID|S_ISVTX|S_IRWXU|S_IRWXG|S_IRWXO)
+#  define ALLPERMS (S_ISUID | S_ISGID | S_ISVTX | S_IRWXU | S_IRWXG | S_IRWXO)
 #endif
 
       ret = fchmodat (destdirfd, de->d_name, mode & ALLPERMS, AT_SYMLINK_NOFOLLOW);
@@ -1800,3 +1873,26 @@ find_annotation (libcrun_container_t *container, const char *name)
 
   return NULL;
 }
+
+ssize_t
+safe_write (int fd, const void *buf, ssize_t count)
+{
+  ssize_t written = 0;
+  if (count < 0)
+    {
+      errno = EINVAL;
+      return -1;
+    }
+  while (written < count)
+    {
+      ssize_t w = write (fd, buf + written, count - written);
+      if (UNLIKELY (w < 0))
+        {
+          if (errno == EINTR || errno == EAGAIN)
+            continue;
+          return w;
+        }
+      written += w;
+    }
+  return written;
+}
