@@ -16,54 +16,137 @@
  * along with crun.  If not, see <http://www.gnu.org/licenses/>.
  */
 #ifndef UTILS_H
-# define UTILS_H
-# include <config.h>
-# include <stdio.h>
-# include <stdlib.h>
-# include <errno.h>
-# include <argp.h>
-# include "error.h"
-# include <dirent.h>
-# include <runtime_spec_schema_config_schema.h>
-# include "container.h"
+#define UTILS_H
 
-# ifndef TEMP_FAILURE_RETRY
-#  define TEMP_FAILURE_RETRY(expression)                                \
-  (__extension__                                                        \
-   ({ long int __result;                                                \
-     do __result = (long int) (expression);                             \
-     while (__result < 0 && errno == EINTR);                            \
-     __result; }))
-# endif
+#include <config.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <errno.h>
+#include <argp.h>
+#include "error.h"
+#include <dirent.h>
+#include <unistd.h>
+#include <runtime_spec_schema_config_schema.h>
+#include "container.h"
 
+#ifndef TEMP_FAILURE_RETRY
+#  define TEMP_FAILURE_RETRY(expression)      \
+    (__extension__({                          \
+      long int __result;                      \
+      do                                      \
+        __result = (long int) (expression);   \
+      while (__result < 0 && errno == EINTR); \
+      __result;                               \
+    }))
+#endif
 
-void cleanup_filep (FILE **f);
-void cleanup_freep (void *p);
-void cleanup_closep (void *p);
-void cleanup_close_vecp (int **p);
-void cleanup_dirp (DIR **p);
-int close_and_reset (int *fd);
+#define cleanup_file __attribute__ ((cleanup (cleanup_filep)))
+#define cleanup_free __attribute__ ((cleanup (cleanup_freep)))
+#define cleanup_close __attribute__ ((cleanup (cleanup_closep)))
+#define cleanup_close_vec __attribute__ ((cleanup (cleanup_close_vecp)))
+#define cleanup_dir __attribute__ ((cleanup (cleanup_dirp)))
+#define arg_unused __attribute__ ((unused))
 
-# define cleanup_file __attribute__((cleanup (cleanup_filep)))
-# define cleanup_free __attribute__((cleanup (cleanup_freep)))
-# define cleanup_close __attribute__((cleanup (cleanup_closep)))
-# define cleanup_close_vec __attribute__((cleanup (cleanup_close_vecp)))
-# define cleanup_dir __attribute__((cleanup (cleanup_dirp)))
-# define arg_unused __attribute__((unused))
+#define LIKELY(x) __builtin_expect ((x), 1)
+#define UNLIKELY(x) __builtin_expect ((x), 0)
 
-# define LIKELY(x) __builtin_expect((x),1)
-# define UNLIKELY(x) __builtin_expect((x),0)
+static inline void *
+xmalloc (size_t size)
+{
+  void *res = malloc (size);
+  if (UNLIKELY (res == NULL))
+    OOM ();
+  return res;
+}
 
-void *xmalloc (size_t size);
+static inline void *
+xmalloc0 (size_t size)
+{
+  void *res = calloc (1, size);
+  if (UNLIKELY (res == NULL))
+    OOM ();
+  return res;
+}
 
-void *xrealloc (void *ptr, size_t size);
+static inline void *
+xrealloc (void *ptr, size_t size)
+{
+  void *res = realloc (ptr, size);
+  if (UNLIKELY (res == NULL))
+    OOM ();
+  return res;
+}
 
-char *xstrdup (const char *str);
+static inline void
+cleanup_freep (void *p)
+{
+  void **pp = (void **) p;
+  free (*pp);
+}
+
+static inline void
+cleanup_filep (FILE **f)
+{
+  FILE *file = *f;
+  if (file)
+    (void) fclose (file);
+}
+
+static inline void
+cleanup_closep (void *p)
+{
+  int *pp = p;
+  if (*pp >= 0)
+    TEMP_FAILURE_RETRY (close (*pp));
+}
+
+static inline void
+cleanup_close_vecp (int **p)
+{
+  int *pp = *p;
+  int i;
+
+  for (i = 0; pp[i] >= 0; i++)
+    TEMP_FAILURE_RETRY (close (pp[i]));
+}
+
+static inline void
+cleanup_dirp (DIR **p)
+{
+  DIR *dir = *p;
+  if (dir)
+    closedir (dir);
+}
+
+static inline int
+close_and_reset (int *fd)
+{
+  int ret = 0;
+  if (*fd >= 0)
+    {
+      ret = TEMP_FAILURE_RETRY (close (*fd));
+      if (LIKELY (ret == 0))
+        *fd = -1;
+    }
+  return ret;
+}
+
+static inline char *
+xstrdup (const char *str)
+{
+  char *ret;
+  if (str == NULL)
+    return NULL;
+
+  ret = strdup (str);
+  if (ret == NULL)
+    OOM ();
+
+  return ret;
+}
 
 int xasprintf (char **str, const char *fmt, ...);
 
-char *argp_mandatory_argument (char *arg, struct argp_state *state);
-
 int crun_path_exists (const char *path, libcrun_error_t *err);
 
 int write_file (const char *name, const void *data, size_t len, libcrun_error_t *err);
@@ -78,11 +161,11 @@ int crun_ensure_directory_at (int dirfd, const char *path, int mode, bool nofoll
 
 int crun_ensure_file_at (int dirfd, const char *path, int mode, bool nofollow, libcrun_error_t *err);
 
-int check_fd_under_path (const char *rootfs, size_t rootfslen, int fd, const char *fdname, libcrun_error_t *err);
+int crun_safe_ensure_directory_at (int dirfd, const char *dirpath, size_t dirpath_len, const char *path, int mode,
+                                   libcrun_error_t *err);
 
-int crun_safe_ensure_directory_at (int dirfd, const char *dirpath, size_t dirpath_len, const char *path, int mode, libcrun_error_t *err);
-
-int crun_safe_ensure_file_at (int dirfd, const char *dirpath, size_t dirpath_len, const char *path, int mode, libcrun_error_t *err);
+int crun_safe_ensure_file_at (int dirfd, const char *dirpath, size_t dirpath_len, const char *path, int mode,
+                              libcrun_error_t *err);
 
 int crun_dir_p (const char *path, bool nofollow, libcrun_error_t *err);
 
@@ -92,8 +175,6 @@ int detach_process ();
 
 int create_file_if_missing_at (int dirfd, const char *file, libcrun_error_t *err);
 
-int create_file_if_missing (const char *file, libcrun_error_t *err);
-
 int check_running_in_user_namespace (libcrun_error_t *err);
 
 int set_selinux_exec_label (const char *label, libcrun_error_t *err);
@@ -106,6 +187,8 @@ int read_all_fd (int fd, const char *description, char **out, size_t *len, libcr
 
 int read_all_file (const char *path, char **out, size_t *len, libcrun_error_t *err);
 
+int read_all_file_at (int dirfd, const char *path, char **out, size_t *len, libcrun_error_t *err);
+
 int open_unix_domain_client_socket (const char *path, int dgram, libcrun_error_t *err);
 
 int open_unix_domain_socket (const char *path, int dgram, libcrun_error_t *err);
@@ -126,7 +209,8 @@ int run_process (char **args, libcrun_error_t *err);
 
 size_t format_default_id_mapping (char **ret, uid_t container_id, uid_t host_id, int is_uid);
 
-int run_process_with_stdin_timeout_envp (char *path, char **args, const char *cwd, int timeout, char **envp, char *stdin, size_t stdin_len, libcrun_error_t *err);
+int run_process_with_stdin_timeout_envp (char *path, char **args, const char *cwd, int timeout, char **envp,
+                                         char *stdin, size_t stdin_len, int out_fd, int err_fd, libcrun_error_t *err);
 
 int close_fds_ge_than (int n, libcrun_error_t *err);
 
@@ -156,6 +240,11 @@ int get_file_type (mode_t *mode, bool nofollow, const char *path);
 
 int get_file_type_fd (int fd, mode_t *mode);
 
-int str2sig (const char *name);
+int safe_openat (int dirfd, const char *rootfs, size_t rootfs_len, const char *path, int flags, int mode,
+                 libcrun_error_t *err);
+
+ssize_t safe_write (int fd, const void *buf, ssize_t count);
+
+LIBCRUN_PUBLIC int libcrun_str2sig (const char *name);
 
 #endif
