@@ -1,7 +1,7 @@
 /*
  * crun - OCI runtime written in C
  *
- * Copyright (C) 2017, 2018, 2019 Giuseppe Scrivano <giuseppe@scrivano.org>
+ * Copyright (C) 2017, 2018, 2019, 2020 Giuseppe Scrivano <giuseppe@scrivano.org>
  * crun is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
@@ -31,34 +31,38 @@
 static char doc[] = "OCI runtime";
 
 enum
-  {
-    OPTION_CONSOLE_SOCKET = 1000,
-    OPTION_PID_FILE,
-    OPTION_NO_SUBREAPER,
-    OPTION_NO_NEW_KEYRING,
-    OPTION_PRESERVE_FDS,
-    OPTION_NO_PIVOT
-  };
+{
+  OPTION_CONSOLE_SOCKET = 1000,
+  OPTION_PID_FILE,
+  OPTION_NO_SUBREAPER,
+  OPTION_NO_NEW_KEYRING,
+  OPTION_PRESERVE_FDS,
+  OPTION_NO_PIVOT
+};
 
 static const char *bundle = NULL;
 
 static libcrun_context_t crun_context;
 
-static struct argp_option options[] =
-  {
-   {"bundle", 'b', 0, 0, "container bundle (default \".\")", 0},
-   {"detach", 'd', 0, 0, "detach from the parent", 0},
-   {"console-socket", OPTION_CONSOLE_SOCKET, "SOCKET", 0, "path to a socket that will receive the master end of the tty", 0},
-   {"preserve-fds", OPTION_PRESERVE_FDS, 0, 0, "pass additional FDs to the container", 0},
-   {"pid-file", OPTION_PID_FILE, "FILE", 0, "where to write the PID of the container", 0},
-   {"no-subreaper", OPTION_NO_SUBREAPER, 0, 0, "do not create a subreaper process", 0},
-   {"no-new-keyring", OPTION_NO_NEW_KEYRING, 0, 0, "keep the same session key", 0},
-   {"no-pivot", OPTION_NO_PIVOT, 0, 0, "do not use pivot_root", 0},
-   { 0, }
-  };
+static struct argp_option options[]
+    = { { "bundle", 'b', "DIR", 0, "container bundle (default \".\")", 0 },
+        { "config", 'f', "FILE", 0, "override the config file name", 0 },
+        { "detach", 'd', 0, 0, "detach from the parent", 0 },
+        { "console-socket", OPTION_CONSOLE_SOCKET, "SOCKET", 0,
+          "path to a socket that will receive the ptmx end of the tty", 0 },
+        { "preserve-fds", OPTION_PRESERVE_FDS, 0, 0, "pass additional FDs to the container", 0 },
+        { "pid-file", OPTION_PID_FILE, "FILE", 0, "where to write the PID of the container", 0 },
+        { "no-subreaper", OPTION_NO_SUBREAPER, 0, 0, "do not create a subreaper process", 0 },
+        { "no-new-keyring", OPTION_NO_NEW_KEYRING, 0, 0, "keep the same session key", 0 },
+        { "no-pivot", OPTION_NO_PIVOT, 0, 0, "do not use pivot_root", 0 },
+        {
+            0,
+        } };
 
 static char args_doc[] = "run [OPTION]... CONTAINER";
 
+static const char *config_file = "config.json";
+
 static error_t
 parse_opt (int key, char *arg, struct argp_state *state)
 {
@@ -68,6 +72,10 @@ parse_opt (int key, char *arg, struct argp_state *state)
       crun_context.detach = true;
       break;
 
+    case 'f':
+      config_file = crun_context.config_file = argp_mandatory_argument (arg, state);
+      break;
+
     case 'b':
       bundle = crun_context.bundle = argp_mandatory_argument (arg, state);
       break;
@@ -114,12 +122,26 @@ crun_command_run (struct crun_global_arguments *global_args, int argc, char **ar
   int first_arg, ret;
   cleanup_free libcrun_container_t *container = NULL;
   cleanup_free char *bundle_cleanup = NULL;
+  cleanup_free char *config_file_cleanup = NULL;
 
   crun_context.preserve_fds = 0;
 
   argp_parse (&run_argp, argc, argv, ARGP_IN_ORDER, &first_arg, &crun_context);
   crun_assert_n_args (argc - first_arg, 1, 1);
 
+  /* Make sure the config is an absolute path before changing the directory.  */
+  if ((strcmp ("config.json", config_file) != 0))
+    {
+      if (config_file[0] != '/')
+        {
+          config_file_cleanup = realpath (config_file, NULL);
+          if (config_file_cleanup == NULL)
+            libcrun_fail_with_error (errno, "realpath `%s` failed", config_file);
+          config_file = config_file_cleanup;
+          crun_context.config_file = config_file;
+        }
+    }
+
   /* Make sure the bundle is an absolute path.  */
   if (bundle)
     {
@@ -135,7 +157,7 @@ crun_command_run (struct crun_global_arguments *global_args, int argc, char **ar
         libcrun_fail_with_error (errno, "chdir `%s` failed", bundle);
     }
 
-  container = libcrun_container_load_from_file ("config.json", err);
+  container = libcrun_container_load_from_file (config_file, err);
   if (container == NULL)
     return -1;
 
